# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer Tar related functions
"""

from freezer.utils import utils


class TarCommandBuilder(object):
    """
    Building a tar cmd command. To build command invoke method build.
    """

    UNIX_TEMPLATE = (
        "{gnutar_path} --create {algo} --warning=none --no-check-device "
        "--one-file-system --preserve-permissions --same-owner "
        "--seek --ignore-failed-read")

    # local windows template
    WINDOWS_TEMPLATE = (
        "{gnutar_path} -c {algo} --incremental --unlink-first "
        "--ignore-zeros")

    LISTED_TEMPLATE = "{tar_command} --listed-incremental={listed_incremental}"

    DEREFERENCE_MODE = {'soft': '--dereference',
                        'hard': '--hard-dereference',
                        'all': '--hard-dereference --dereference'}

    def __init__(self, filepath, compression_algo, is_windows, tar_path=None):
        self.tar_path = tar_path or utils.tar_path()
        self.dereference = ''
        self.listed_incremental = None
        self.exclude = ''
        self.openssl_path = None
        self.encrypt_pass_file = None
        self.output_file = None
        self.filepath = filepath
        self.compression_algo = get_tar_flag_from_algo(compression_algo)
        self.is_windows = is_windows

    def set_listed_incremental(self, absolute_path):
        self.listed_incremental = absolute_path

    def set_exclude(self, exclude):
        self.exclude = exclude

    def set_dereference(self, mode):
        """
        Dereference hard and soft links according option choices.
            'soft' dereference soft links,
            'hard' dereference hardlinks,
            'all' dereference both.
            Default 'None'.
        """
        self.dereference = self.DEREFERENCE_MODE[mode]

    def set_encryption(self, encrypt_pass_file, openssl_path=None):
        self.openssl_path = openssl_path or utils.openssl_path()
        self.encrypt_pass_file = encrypt_pass_file

    def build(self):
        if self.is_windows:
            tar_command = self.WINDOWS_TEMPLATE.format(
                gnutar_path=self.tar_path, algo=self.compression_algo)
        else:
            tar_command = self.UNIX_TEMPLATE.format(
                gnutar_path=self.tar_path, algo=self.compression_algo)

        if self.dereference:
            tar_command = "{0} {1}".format(tar_command, self.dereference)

        if self.listed_incremental:
            tar_command = self.LISTED_TEMPLATE.format(
                tar_command=tar_command,
                listed_incremental=self.listed_incremental)

        if self.exclude:
            tar_command = '{tar_command} --exclude="{exclude}"'.format(
                tar_command=tar_command, exclude=self.exclude)

        tar_command = '{0} {1}'.format(tar_command, self.filepath)

        if self.encrypt_pass_file:
            openssl_cmd = "{openssl_path} enc -aes-256-cfb -pass file:{file}"\
                .format(openssl_path=self.openssl_path,
                        file=self.encrypt_pass_file)
            tar_command = '{0} | {1} && exit ${{PIPESTATUS[0]}}'\
                .format(tar_command, openssl_cmd)

        return tar_command


class TarCommandRestoreBuilder(object):
    WINDOWS_TEMPLATE = '{0} -x {1} --incremental --unlink-first ' \
        '--ignore-zeros'
    DRY_RUN_TEMPLATE = '{0} {1} --incremental --list ' \
        '--ignore-zeros --warning=none'
    UNIX_TEMPLATE = '{0} {1} --incremental --extract ' \
        '--ignore-zeros --warning=none --overwrite --directory {2}'
    OPENSSL_DEC = "{openssl_path} enc -d -aes-256-cfb -pass file:{file}"

    def __init__(self, restore_path, compression_algo, is_windows,
                 tar_path=None):
        self.dry_run = False
        self.is_windows = False
        self.openssl_path = None
        self.encrypt_pass_file = None
        self.tar_path = tar_path or utils.tar_path()
        self.restore_path = restore_path
        self.compression_algo = get_tar_flag_from_algo(compression_algo)
        self.is_windows = is_windows

    def set_dry_run(self):
        self.dry_run = True

    def set_encryption(self, encrypt_pass_file, openssl_path=None):
        self.openssl_path = openssl_path or utils.openssl_path()
        self.encrypt_pass_file = encrypt_pass_file

    def build(self):
        if self.dry_run:
            tar_command = self.DRY_RUN_TEMPLATE.format(self.tar_path,
                                                       self.compression_algo)
        elif self.is_windows:
            tar_command = self.WINDOWS_TEMPLATE.format(self.tar_path,
                                                       self.compression_algo)
        else:
            tar_command = self.UNIX_TEMPLATE.format(self.tar_path,
                                                    self.compression_algo,
                                                    self.restore_path)

        # Check if encryption file is provided and set the openssl decrypt
        # command accordingly
        if self.encrypt_pass_file:
            openssl_cmd = self.OPENSSL_DEC.format(
                openssl_path=self.openssl_path,
                file=self.encrypt_pass_file)
            tar_command = '{0} | {1} && exit ${{PIPESTATUS[0]}}'\
                .format(openssl_cmd, tar_command)
        return tar_command


def get_tar_flag_from_algo(compression):
    algo = {
        'gzip': '-z',
        'bzip2': '-j',
        'xz': '-J',
    }
    compression_exec = utils.get_executable_path(compression)
    if not compression_exec:
        raise Exception("Critical Error: {0} executable not found ".
                        format(compression))
    return algo.get(compression)



# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import distutils.spawn
import hashlib
import itertools
import os
import random
import shutil
import subprocess
import tempfile
import unittest

from oslo_serialization import jsonutils as json

import paramiko


FREEZERC = distutils.spawn.find_executable('freezer-agent')


class CommandFailed(Exception):
    def __init__(self, returncode, cmd, output, stderr):
        super(CommandFailed, self).__init__()
        self.returncode = returncode
        self.cmd = cmd
        self.stdout = output
        self.stderr = stderr

    def __str__(self):
        return ("Command '%s' returned unexpected exit status %d.\n"
                "stdout:\n%s\n"
                "stderr:\n%s" % (self.cmd, self.returncode,
                                 self.stdout, self.stderr))


def dict_to_args(d):
    ll = [['--' + k.replace('_', '-'), v] for k, v in d.items()]
    return list(itertools.chain.from_iterable(ll))


def execute_freezerc(dict, must_fail=False, merge_stderr=False):
    """
    :param dict:
    :type dict: dict[str, str]
    :param must_fail:
    :param merge_stderr:
    :return:
    """
    return execute([FREEZERC] + dict_to_args(dict), must_fail=must_fail,
                   merge_stderr=merge_stderr)


def execute(args, must_fail=False, merge_stderr=False):
    """
    Executes specified command for the given action.
    :param args:
    :type args: list[str]
    :param must_fail:
    :param merge_stderr:
    :return:
    """
    stdout = subprocess.PIPE
    stderr = subprocess.STDOUT if merge_stderr else subprocess.PIPE
    proc = subprocess.Popen(args,
                            stdout=stdout,
                            stderr=stderr, universal_newlines=True)
    result, result_err = proc.communicate()

    if not must_fail and proc.returncode != 0:
        raise CommandFailed(proc.returncode, ' '.join(args), result,
                            result_err)
    if must_fail and proc.returncode == 0:
        raise CommandFailed(proc.returncode, ' '.join(args), result,
                            result_err)
    return result


class Temp_Tree(object):
    def __init__(self, suffix='', dir=None, create=True):
        self.create = create
        if create:
            self.path = tempfile.mkdtemp(dir=dir, prefix='__freezer_',
                                         suffix=suffix)
        else:
            self.path = dir
        self.files = []

    def __enter__(self):
        return self

    def cleanup(self):
        if self.create and self.path:
            shutil.rmtree(self.path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def add_random_data(self, ndir=5, nfile=5, size=1024):
        """
        add some files containing random data

        :param ndir: number of dirs to create
        :param nfile: number of files to create in each dir
        :param size: size of files
        :return: None
        """
        def create_file(path):
            abs_pathname = self.create_file_with_random_data(
                dir_path=path, size=size)
            rel_path_name = abs_pathname[len(self.path) + 1:]
            self.files.append(rel_path_name)

        for _ in range(nfile):
            create_file(self.path)

        for _ in range(ndir):
            subdir_path = tempfile.mkdtemp(dir=self.path)
            for _ in range(nfile):
                create_file(subdir_path)

    def create_file_with_random_data(self, dir_path, size=1024):
        handle, abs_pathname = tempfile.mkstemp(dir=dir_path)
        with open(abs_pathname, 'wb') as fd:
            fd.write(os.urandom(size))
        return abs_pathname

    def modify_random_files(self, count=1):
        indexes = []
        for _ in range(count):
            indexes.append(random.randint(0, len(self.files) - 1))
        for file_index in indexes:
            file_name = self.files[file_index]
            with open(os.path.join(self.path, file_name), 'ab') as fd:
                size_to_add = int(fd.tell() * 0.5)
                fd.write(os.urandom(size_to_add))

    def delete_random_files(self, count=1):
        indexes = []
        for _ in range(count):
            indexes.append(random.randint(0, len(self.files) - 1))
        for file_index in indexes:
            file_name = self.files[file_index]
            os.unlink(os.path.join(self.path, file_name))

    def get_file_hash(self, rel_filepath):
        filepath = os.path.join(self.path, rel_filepath)
        if os.path.isfile(filepath):
            return self._filehash(filepath)
        else:
            return ''

    def _filehash(self, filepath):
        """
        Get GIT style sha1 hash for a file

        :param filepath: path of file to hash
        :return: hash of the file
        """
        filesize_bytes = os.path.getsize(filepath)
        hash_obj = hashlib.sha1()
        hash_obj.update(("blob %u\0" % filesize_bytes).encode('utf-8'))
        with open(filepath, 'rb') as handle:
            hash_obj.update(handle.read())
        return hash_obj.hexdigest()

    def get_file_list(self):
        """
        walks the dir tree and creates a list of relative pathnames
        :return: list of relative file paths
        """
        self.files = []
        for root, dirs, files in os.walk(self.path):
            rel_base = root[len(self.path) + 1:]
            self.files.extend([os.path.join(rel_base, x) for x in files])
        return self.files

    def is_equal(self, other_tree):
        """
        Checks whether two dir tree contain the same files
        It checks the number of files and the hash of each file.

        NOTE: tox puts .coverage files in the temp folder (?)

        :param other_tree: dir tree to compare with
        :return: true if the dir trees contain the same files
        """
        lh_files = [x for x in sorted(self.get_file_list())
                    if not x.startswith('.coverage')]
        rh_files = [x for x in sorted(other_tree.get_file_list())
                    if not x.startswith('.coverage')]
        if lh_files != rh_files:
            return False
        for fname in lh_files:
            if os.path.isfile(fname):
                if self.get_file_hash(fname) != \
                        other_tree.get_file_hash(fname):
                    return False
        return True


class TestFS(unittest.TestCase):
    """
    Utility class for setting up the tests.

    Type of tests depends (also) on the environment variables defined.

    To enable the ssh storage testing, the following environment
    variables need to be defined:
    - FREEZER_TEST_SSH_KEY
    - FREEZER_TEST_SSH_USERNAME
    - FREEZER_TEST_SSH_HOST
    - FREEZER_TEST_CONTAINER

    To enable the swift storage testing, the following environment
    variables need to be defined:
    - FREEZER_TEST_OS_TENANT_NAME
    - FREEZER_TEST_OS_USERNAME
    - FREEZER_TEST_OS_REGION_NAME
    - FREEZER_TEST_OS_PASSWORD
    - FREEZER_TEST_OS_AUTH_URL

    Tests involving LVM snapshots are evoided if:
    - user is not root
    - FREEZER_TEST_NO_LVM is set
    """

    ssh_key = os.environ.get('FREEZER_TEST_SSH_KEY')
    ssh_username = os.environ.get('FREEZER_TEST_SSH_USERNAME')
    ssh_host = os.environ.get('FREEZER_TEST_SSH_HOST')
    container = os.environ.get('FREEZER_TEST_CONTAINER')
    use_ssh = ssh_key and ssh_username and ssh_host and container

    os_project_name = os.environ.get('FREEZER_TEST_OS_TENANT_NAME')
    os_user_name = os.environ.get('FREEZER_TEST_OS_USERNAME')
    os_region = os.environ.get('FREEZER_TEST_OS_REGION_NAME')
    os_password = os.environ.get('FREEZER_TEST_OS_PASSWORD')
    os_auth_url = os.environ.get('FREEZER_TEST_OS_AUTH_URL')
    use_os = (os_project_name and os_user_name and os_region and
              os_password and os_auth_url)
    if use_os:
        os.environ['OS_USERNAME'] = os_user_name
        os.environ['OS_PROJECT_NAME'] = os_project_name
        os.environ['OS_AUTH_URL'] = os_auth_url
        os.environ['OS_PASSWORD'] = os_password
        os.environ['OS_REGION_NAME'] = os_region
        os.environ['OS_PROJECT_ID'] = ''

    openstack_executable = distutils.spawn.find_executable('openstack')
    swift_executable = distutils.spawn.find_executable('swift')

    use_lvm = (os.getuid() == 0 and 'FREEZER_TEST_NO_LVM' not in os.environ)
    ssh_executable = distutils.spawn.find_executable('ssh')

    def setUp(self):
        self.source_tree = Temp_Tree()
        self.dest_tree = Temp_Tree()
        if TestFS.use_ssh:
            self.ssh_client = paramiko.SSHClient()
            self.ssh_client.set_missing_host_key_policy(
                paramiko.AutoAddPolicy())
            self.ssh_client.connect(TestFS.ssh_host,
                                    username=TestFS.ssh_username,
                                    key_filename=TestFS.ssh_key)

    def tearDown(self):
        self.source_tree.cleanup()
        self.dest_tree.cleanup()

    def assertTreesMatch(self):
        self.assertTrue(self.source_tree.is_equal(self.dest_tree))

    def assertTreesMatchNot(self):
        self.assertFalse(self.source_tree.is_equal(self.dest_tree))

    def get_file_list_ssh(self, sub_path=''):
        ftp = self.ssh_client.open_sftp()
        path = '{0}/{1}'.format(self.container, sub_path)
        return ftp.listdir(path)

    def remove_ssh_directory(self, sub_path=''):
        cmd = 'rm -rf {0}/{1}'.format(self.container, sub_path)
        self.ssh_client.exec_command(cmd)

    def get_file_list_openstack(self, container):
        if self.openstack_executable:
            json_result = execute([self.openstack_executable, 'object', 'list',
                                   container, '-f', 'json'])
            result = json.loads(json_result)
            return [x['Name'] for x in result]
        if self.swift_executable:
            result = execute([self.swift_executable, 'list', container])
            return result.split()
        raise Exception(
            "Unable to get container list using openstackclient/swiftclient")

    def remove_swift_container(self, container):
        if self.openstack_executable:
            execute([self.openstack_executable, 'container',
                     'delete', container])
            execute([self.openstack_executable, 'container',
                     'delete', container + '_segments'])
        elif self.swift_executable:
            execute([self.swift_executable, 'delete', container])
            execute([self.swift_executable, 'delete', container + '_segments'])
        return True

    def do_backup_and_restore_with_check(self, backup_args, restore_args):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()
        result = execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()



# (c) Copyright 2018 ZTE Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import tempfile
import unittest

# from unittest import mock
# from mock import patch

from freezer.utils import crypt


class AESCipherTestCase(unittest.TestCase):

    def setUp(self):
        super(AESCipherTestCase, self).setUp()
        self.passwd_test_file_dir = None
        self.passwd_test_file_name = None
        self.create_pass_test_file()

    def tearDown(self):
        super(AESCipherTestCase, self).tearDown()
        self.delete_pass_test_file()

    def create_pass_test_file(self):
        if self.passwd_test_file_name:
            return
        tmpdir = tempfile.mkdtemp()
        FILES_DIR_PREFIX = "freezer_passwd_dir"
        files_dir = tempfile.mkdtemp(dir=tmpdir, prefix=FILES_DIR_PREFIX)
        file_name = "passwd_test"
        self.passwd_test_file_dir = files_dir
        text = '78f40f2c57eee727a4be179049cecf89'
        filehandle = open(files_dir + "/" + file_name, 'w')
        if filehandle:
            filehandle.write(text)
            filehandle.close()
            self.passwd_test_file_name = files_dir + "/" + file_name

    def delete_pass_test_file(self):
        if self.passwd_test_file_name:
            files_dir = self.passwd_test_file_dir
            shutil.rmtree(files_dir)
            self.passwd_test_file_name = None
            self.passwd_test_file_dir = None

    def test_get_pass_from_file(self):
        pfile = self.passwd_test_file_name
        passwd = crypt.AESCipher._get_pass_from_file(pfile)
        self.assertEqual(passwd, '78f40f2c57eee727a4be179049cecf89')

#    @unittest.skipIf(sys.version_info.major == 3,
#                     'Not supported on python v 3.x')
    def test_derive_key_and_iv(self):
        passwd = b'ababab'
        salt = b'a'
        ret1, ret2 = crypt.AESCipher._derive_key_and_iv(password=passwd,
                                                        salt=salt,
                                                        key_length=10,
                                                        iv_length=5)
        expect1 = b'\xb3J5\xce\xd4b\x87\xce\xe0:'
        expect2 = b'\x93\xc9\x9d\x03\x00'
        self.assertEqual(ret1, expect1)
        self.assertEqual(ret2, expect2)



# (C) Copyright 2017 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

import ddt

from oslo_serialization import jsonutils as json

from freezer.engine.nova import nova
from freezer.tests import commons


class FakeServer(object):
    def __init__(self, instance_id):
        self.id = instance_id


class FakeContextManager(object):
    def __init__(self, mock_file):
        self.mock_file = mock_file

    def __enter__(self):
        return self.mock_file

    def __exit__(self, *args):
        pass


class TestNovaEngine(commons.FreezerBaseTestCase):

    def setUp(self):
        super(TestNovaEngine, self).setUp()
        self.backup_opt = commons.BackupOpt1()
        self.project_id = "test-project-id"
        self.instance_names = ["instance-1", "instance-2", "instance-3"]
        if self.instance_names:
            self.instance_ids = [server for server in
                                 self.instance_names
                                 if server == self.backup_opt.nova_inst_name]
        self.instance_ids = ["instance-id-1", "instance-id-2", "instance-id-3"]

        self.instance_ids_str = json.dumps(self.instance_ids)
        servers_list = [FakeServer(instance_id) for instance_id in
                        self.instance_ids]

        self.mock_nova = mock.MagicMock()
        self.mock_nova.servers.list = mock.Mock(return_value=servers_list)
        self.backup_opt.client_manager = mock.MagicMock()
        self.backup_opt.client_manager.get_nova.return_value = self.mock_nova
        self.expected_backup_calls = [
            mock.call(
                backup_resource=inst_id,
                hostname_backup_name=os.path.join(self.backup_opt.backup_name,
                                                  inst_id),
                no_incremental=self.backup_opt.no_incremental,
                max_level=self.backup_opt.max_level,
                always_level=self.backup_opt.always_level,
                restart_always_level=self.backup_opt.restart_always_level)
            for inst_id in self.instance_ids]
        self.expected_restore_calls = [
            mock.call(
                hostname_backup_name=os.path.join(self.backup_opt.backup_name,
                                                  inst_id),
                restore_resource=inst_id,
                overwrite=self.backup_opt.overwrite,
                recent_to_date='test_timestamp')
            for inst_id in self.instance_ids]


class TestNovaEngineSwiftStorage(TestNovaEngine):
    def setUp(self):
        super(TestNovaEngineSwiftStorage, self).setUp()
        self.mock_swift_connection = mock.MagicMock()
        self.mock_swift_connection.get_object.return_value = (
            None, self.instance_ids_str)
        self.mock_swift_connection.put_object = mock.MagicMock()

        self.mock_swift_storage = mock.MagicMock()
        self.mock_swift_storage._type = 'swift'

        self.engine = nova.NovaEngine(self.mock_swift_storage)
        self.engine.client = self.backup_opt.client_manager
        self.engine.client.create_swift = mock.Mock(
            return_value=self.mock_swift_connection)
        self.engine.backup = mock.Mock()
        self.engine.restore = mock.Mock()
        self.engine.nova = self.mock_nova

    def test_backup_nova_tenant_to_swift_storage(self):
        self.engine.backup_nova_tenant(self.project_id,
                                       self.backup_opt.backup_name,
                                       self.backup_opt.no_incremental,
                                       self.backup_opt.max_level,
                                       self.backup_opt.always_level,
                                       self.backup_opt.restart_always_level)

        self.mock_nova.servers.list.assert_called_once_with(detailed=False)
        self.engine.client.create_swift.assert_called_once()
        self.mock_swift_connection.put_object.assert_called_with(
            self.mock_swift_storage.storage_path,
            "project_test-project-id",
            self.instance_ids_str)
        self.engine.backup.assert_has_calls(self.expected_backup_calls,
                                            any_order=True)

    def test_restore_nova_tenant_from_swift_storage(self):
        self.engine.restore_nova_tenant(self.project_id,
                                        self.backup_opt.backup_name,
                                        self.backup_opt.overwrite,
                                        'test_timestamp')

        self.engine.client.create_swift.assert_called_once()
        self.mock_swift_connection.get_object.assert_called_with(
            self.mock_swift_storage.storage_path,
            "project_test-project-id")
        self.engine.restore.assert_has_calls(self.expected_restore_calls,
                                             any_order=True)


@ddt.ddt
class TestNovaEngineFSLikeStorage(TestNovaEngine):
    def setUp(self):
        super(TestNovaEngineFSLikeStorage, self).setUp()
        self.mock_file = mock.Mock()
        self.mock_file.readline = mock.Mock(return_value=self.instance_ids_str)
        self.mock_file.write = mock.Mock()

        self.mock_fslike_storage = mock.MagicMock()
        self.mock_fslike_storage.open = mock.Mock(
            return_value=FakeContextManager(self.mock_file))
        self.mock_fslike_storage.storage_path = 'test/storage/path'
        self.local_backup_file = os.path.join(
            self.mock_fslike_storage.storage_path,
            "project_test-project-id")

        self.engine = nova.NovaEngine(self.mock_fslike_storage)
        self.engine.client = self.backup_opt.client_manager
        self.engine.backup = mock.Mock()
        self.engine.restore = mock.Mock()
        self.engine.nova = self.mock_nova

    @ddt.data('local', 'ssh')
    def test_backup_nova_tenant_to_fslike_storage(self,
                                                  storage_type):
        self.mock_fslike_storage._type = storage_type
        self.engine.backup_nova_tenant(self.project_id,
                                       self.backup_opt.backup_name,
                                       self.backup_opt.no_incremental,
                                       self.backup_opt.max_level,
                                       self.backup_opt.always_level,
                                       self.backup_opt.restart_always_level)

        self.mock_nova.servers.list.assert_called_once_with(detailed=False)
        self.mock_fslike_storage.open.assert_called_once_with(
            self.local_backup_file,
            'wb')
        self.mock_file.write.assert_called_once_with(self.instance_ids_str)
        self.engine.backup.assert_has_calls(self.expected_backup_calls,
                                            any_order=True)

    @ddt.data('local', 'ssh')
    def test_restore_nova_tenant_from_fslike_storage(self,
                                                     storage_type):
        self.mock_fslike_storage._type = storage_type
        self.engine.restore_nova_tenant(self.project_id,
                                        self.backup_opt.backup_name,
                                        self.backup_opt.overwrite,
                                        'test_timestamp')

        self.mock_fslike_storage.open.assert_called_once_with(
            self.local_backup_file,
            'rb')
        self.mock_file.readline.assert_called_once_with()
        self.engine.restore.assert_has_calls(self.expected_restore_calls,
                                             any_order=True)


class TestNovaEngineS3Storage(TestNovaEngine):
    def setUp(self):
        super(TestNovaEngineS3Storage, self).setUp()
        self.mock_s3_storage = mock.MagicMock()
        self.mock_s3_storage.get_object()['Body'].read.return_value \
            = self.instance_ids_str
        self.mock_s3_storage.put_object = mock.MagicMock()
        self.mock_s3_storage.storage_path = 'test/storage/path'

        self.mock_s3_storage._type = 's3'

        self.engine = nova.NovaEngine(self.mock_s3_storage)
        self.engine.client = self.backup_opt.client_manager
        self.engine.backup = mock.Mock()
        self.engine.restore = mock.Mock()
        self.engine.nova = self.mock_nova

    def test_backup_nova_tenant_to_s3_storage(self):
        self.engine.backup_nova_tenant(self.project_id,
                                       self.backup_opt.backup_name,
                                       self.backup_opt.no_incremental,
                                       self.backup_opt.max_level,
                                       self.backup_opt.always_level,
                                       self.backup_opt.restart_always_level)

        self.mock_nova.servers.list.assert_called_once_with(detailed=False)
        self.mock_s3_storage.put_object.assert_called_with(
            bucket_name=self.mock_s3_storage.get_bucket_name(),
            key="{0}/project_test-project-id".format(
                self.mock_s3_storage.get_object_prefix()),
            data=self.instance_ids_str
        )
        self.engine.backup.assert_has_calls(self.expected_backup_calls,
                                            any_order=True)

    def test_restore_nova_tenant_from_s3_storage(self):
        self.engine.restore_nova_tenant(self.project_id,
                                        self.backup_opt.backup_name,
                                        self.backup_opt.overwrite,
                                        'test_timestamp')

        self.mock_s3_storage.get_object.assert_called_with(
            bucket_name=self.mock_s3_storage.get_bucket_name(),
            key="{0}/project_test-project-id".format(
                self.mock_s3_storage.get_object_prefix()
            )
        )
        self.engine.restore.assert_has_calls(self.expected_restore_calls,
                                             any_order=True)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer general utils functions
"""

import os
import subprocess

from oslo_log import log

from freezer.engine import engine
from freezer.engine.tar import tar_builders
from freezer.utils import winutils

LOG = log.getLogger(__name__)


class TarEngine(engine.BackupEngine):

    def __init__(
            self, compression, symlinks, exclude, storage,
            max_segment_size, encrypt_key=None,
            dry_run=False, **kwargs):
        """
            :type storage: freezer.storage.base.Storage
        :return:
        """
        self.compression_algo = compression
        self.encrypt_pass_file = encrypt_key
        self.dereference_symlink = symlinks
        self.exclude = exclude
        self.storage = storage
        self.is_windows = winutils.is_windows()
        self.dry_run = dry_run
        self.max_segment_size = max_segment_size
        super(TarEngine, self).__init__(storage=storage)

    @property
    def name(self):
        return "tar"

    def metadata(self, *args):
        return {
            "engine_name": self.name,
            "compression": self.compression_algo,
            # the encrypt_pass_file might be key content so we need to covert
            # to boolean
            "encryption": bool(self.encrypt_pass_file)
        }

    def backup_data(self, backup_resource, manifest_path):
        LOG.info("Starting Tar engine backup stream")
        tar_command = tar_builders.TarCommandBuilder(
            backup_resource, self.compression_algo, self.is_windows)
        if self.encrypt_pass_file:
            tar_command.set_encryption(self.encrypt_pass_file)
        if self.dereference_symlink:
            tar_command.set_dereference(self.dereference_symlink)
        tar_command.set_exclude(self.exclude)
        tar_command.set_listed_incremental(manifest_path)

        command = tar_command.build()

        LOG.info("Execution command: \n{}".format(command))

        tar_process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            shell=True, executable='/bin/bash')
        read_pipe = tar_process.stdout
        tar_chunk = read_pipe.read(self.max_segment_size)
        while tar_chunk:
            yield tar_chunk
            tar_chunk = read_pipe.read(self.max_segment_size)

        self.check_process_output(tar_process, 'Backup')

        LOG.info("Tar engine stream completed")

    def restore_level(self, restore_resource, read_pipe, backup, except_queue):
        """
        Restore the provided file into backup_opt_dict.restore_abs_path
        Decrypt the file if backup_opt_dict.encrypt_pass_file key is provided

        :param restore_path:
        :param read_pipe:
        :type backup: freezer.storage.base.Backup
        :param backup:
        """

        try:

            metadata = backup.metadata()
            if (not self.encrypt_pass_file and
                    metadata.get("encryption", False)):
                raise Exception("Cannot restore encrypted backup without key")

            tar_command = tar_builders.TarCommandRestoreBuilder(
                restore_resource,
                metadata.get('compression', self.compression_algo),
                self.is_windows)

            if self.encrypt_pass_file:
                tar_command.set_encryption(self.encrypt_pass_file)

            if self.dry_run:
                tar_command.set_dry_run()

            command = tar_command.build()

            if winutils.is_windows():
                # on windows, chdir to restore path.
                os.chdir(restore_resource)

            tar_process = subprocess.Popen(
                command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, shell=True, executable='/bin/bash')
            # Start loop reading the pipe and pass the data to the tar
            # std input. If EOFError exception is raised, the loop end
            # the std err will be checked for errors.
            try:
                while True:
                    tar_process.stdin.write(read_pipe.recv_bytes())
            except EOFError:
                LOG.info('Pipe closed as EOF reached. '
                         'Data transmitted successfully')
            finally:
                self.check_process_output(tar_process, 'Restore')

        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    @staticmethod
    def check_process_output(process, function):

        """Check process stderr and return code to determine if error occurred.

        Check process stderr for any non-empty value.
        Check process return code for any non-zero value.
        Log error message and raise an exception if error occurred.

        :param process: the multiprocessing process to check
        :param function: a string: ('Restore' | 'Backup') for error message
        :returns: None -- Do nothing and return None if no error occurred
        :raises: Exception -- Raise Exception if error occurred
        """
        tar_err = process.communicate()[1]

        if tar_err:
            LOG.error('{0} error: {1}'.format(function, tar_err))

        if process.returncode:
            LOG.error('{0} return code is not 0'
                      .format(process.returncode))
            raise Exception('{0} process failed with return code: {1}'
                            .format(function, process.returncode))



# (c) Copyright 2020 ZTE Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue
import unittest

from freezer.utils import streaming


class StreamingTestCase(unittest.TestCase):

    def create_fifo(self, size):
        input_queue = streaming.RichQueue(size)
        read_except_queue = queue.Queue()
        write_except_queue = queue.Queue()

        read_stream = streaming.QueuedThread(self.backup_stream,
                                             input_queue,
                                             read_except_queue)

        write_stream = streaming.QueuedThread(self.write_stream,
                                              input_queue,
                                              write_except_queue)

        read_stream.daemon = True
        write_stream.daemon = True
        read_stream.start()
        write_stream.start()
        read_stream.join()
        write_stream.join()

        return input_queue

    def backup_stream(self, rich_queue):
        rich_queue.put_messages('ancd')

    def write_stream(self, rich_queue):
        for message in rich_queue.get_messages():
            pass

    def test_stream_1(self):
        input_queue = self.create_fifo(1)
        assert input_queue.finish_transmission is True

    def test_stream_2(self):
        input_queue = self.create_fifo(2)
        assert input_queue.finish_transmission is True

    def test_stream_3(self):
        input_queue = self.create_fifo(3)
        assert input_queue.finish_transmission is True

    def test_stream_4(self):
        input_queue = self.create_fifo(4)
        assert input_queue.finish_transmission is True

    def test_stream_5(self):
        input_queue = self.create_fifo(5)
        assert input_queue.finish_transmission is True




# Copyright 2011-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import socket

from cinderclient import exceptions
from oslo_concurrency import processutils


def get_my_ip():
    try:
        csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        csock.connect(('8.8.8.8', 80))
        (addr, port) = csock.getsockname()
        csock.close()
        return addr
    except socket.error:
        return None


def get_root_helper():
    # NOTE (e0ne): We don't use rootwrap now
    return 'sudo'


def require_root(f):
    def wrapper(*args, **kwargs):
        if os.getuid() != 0:
            raise exceptions.CommandError(
                "This command requires root permissions.")
        return f(*args, **kwargs)
    return wrapper


def safe_execute(cmd):
    try:
        processutils.execute(*cmd, root_helper=get_root_helper(),
                             run_as_root=True)
    except processutils.ProcessExecutionError as e:
        print('Command "{0}" execution returned {1} exit code:'.format(
              e.cmd, e.exit_code))
        print('Stderr: {0}'.format(e.stderr))
        print('Stdout: {0}'.format(e.stdout))



# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import configparser
import datetime
import os
import subprocess
import tempfile
import time

from freezer.utils import utils
from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils as json


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class StopState(object):

    @staticmethod
    def stop(job, doc):
        job.job_doc = doc
        job.event = Job.NO_EVENT
        job.job_doc_status = Job.STOP_STATUS
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def abort(job, doc):
        return StopState.stop(job, doc)

    @staticmethod
    def start(job, doc):
        job.job_doc = doc
        job.event = Job.NO_EVENT
        job.job_doc_status = Job.STOP_STATUS
        job.schedule()
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.unschedule()
        job.job_doc_status = Job.REMOVED_STATUS
        return Job.NO_EVENT


class ScheduledState(object):

    @staticmethod
    def stop(job, doc):
        job.unschedule()
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.STOP_EVENT

    @staticmethod
    def abort(job, doc):
        return StopState.stop(job, doc)

    @staticmethod
    def start(job, doc):
        job.event = Job.NO_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.unschedule()
        job.job_doc_status = Job.REMOVED_STATUS
        return Job.NO_EVENT


class RunningState(object):

    @staticmethod
    def stop(job, doc):
        job.event = Job.STOP_EVENT
        return Job.NO_EVENT

    @staticmethod
    def abort(job, doc):
        job.event = Job.ABORT_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.ABORTED_RESULT

    @staticmethod
    def start(job, doc):
        job.event = Job.NO_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.event = Job.REMOVE_EVENT
        return Job.NO_EVENT


class Job(object):

    NO_EVENT = ''
    STOP_EVENT = 'stop'
    START_EVENT = 'start'
    ABORT_EVENT = 'abort'
    REMOVE_EVENT = 'remove'

    STOP_STATUS = 'stop'
    SCHEDULED_STATUS = 'scheduled'
    RUNNING_STATUS = 'running'
    REMOVED_STATUS = 'removed'
    COMPLETED_STATUS = 'completed'

    FAIL_RESULT = 'fail'
    SUCCESS_RESULT = 'success'
    ABORTED_RESULT = 'aborted'

    TIME_NULL = -1

    @staticmethod
    def create(scheduler, executable, job_doc):
        job = Job(scheduler, executable, job_doc)
        if job.job_doc_status in ['running', 'scheduled']:
            LOG.warning('Resetting {0} status from job {1}'
                        .format(job.job_doc_status, job.id))
        if job.job_doc_status == 'stop' and not job.event:
            LOG.info('Job {0} was stopped.'.format(job.id))
            job.event = Job.STOP_EVENT
        elif not job.event:
            LOG.info('Autostart Job {0}'.format(job.id))
            job.event = Job.START_EVENT
        return job

    def __init__(self, scheduler, executable, job_doc):
        self.scheduler = scheduler
        self.executable = executable
        self.job_doc = job_doc
        self.process = None
        self.state = StopState

    def remove(self):
        with self.scheduler.lock:
            # delegate to state object
            LOG.info('REMOVE job {0}'.format(self.id))
            self.state.remove(self)

    @property
    def id(self):
        return self.job_doc['job_id']

    @property
    def session_id(self):
        return self.job_doc.get('session_id', '')

    @session_id.setter
    def session_id(self, value):
        self.job_doc['session_id'] = value

    @property
    def session_tag(self):
        return self.job_doc.get('session_tag', 0)

    @session_tag.setter
    def session_tag(self, value):
        self.job_doc['session_tag'] = value

    @property
    def event(self):
        return self.job_doc['job_schedule'].get('event', '')

    @event.setter
    def event(self, value):
        self.job_doc['job_schedule']['event'] = value

    @property
    def job_doc_status(self):
        return self.job_doc['job_schedule'].get('status', '')

    @job_doc_status.setter
    def job_doc_status(self, value):
        self.job_doc['job_schedule']['status'] = value

    @property
    def result(self):
        return self.job_doc['job_schedule'].get('result', '')

    @result.setter
    def result(self, value):
        self.job_doc['job_schedule']['result'] = value

    def can_be_removed(self):
        return self.job_doc_status == Job.REMOVED_STATUS

    @staticmethod
    def save_action_to_file(action, f):
        parser = configparser.ConfigParser()
        parser.add_section('action')
        for action_k, action_v in action.items():
            if action_v is not None:
                parser.set('action', action_k, str(action_v))
        parser.write(f)
        f.seek(0)

    @property
    def schedule_date(self):
        return self.job_doc['job_schedule'].get('schedule_date', '')

    @property
    def schedule_interval(self):
        return self.job_doc['job_schedule'].get('schedule_interval', '')

    @property
    def schedule_start_date(self):
        return self.job_doc['job_schedule'].get('schedule_start_date', '')

    @property
    def schedule_end_date(self):
        return self.job_doc['job_schedule'].get('schedule_end_date', '')

    @property
    def schedule_cron_fields(self):
        cron_fields = ['year', 'month', 'day', 'week', 'day_of_week',
                       'hour', 'minute', 'second']
        cron_schedule = {}
        for cron in self.job_doc['job_schedule'].keys():
            if cron.startswith('schedule_'):
                cron_key = cron.split('_', 1)[1]
                cron_schedule.update({
                    cron_key: self.job_doc['job_schedule'][cron]})
        return {key: value
                for key, value in cron_schedule.items()
                if key in cron_fields}

    @property
    def scheduled(self):
        return self.scheduler.is_scheduled(self.id)

    def get_schedule_args(self):
        def get_start_date(date):
            # start_date format "%Y-%m-%dT%H:%M:%S"
            now = datetime.datetime.now()
            start_date = now + datetime.timedelta(0, 2, 0)
            if (utils.date_to_timestamp(date) >
                    utils.date_to_timestamp(now.isoformat().split('.')[0])):
                start_date = datetime.datetime.strptime(
                    date, "%Y-%m-%dT%H:%M:%S")
            return start_date

        def get_end_date(start, end):
            # start end format "%Y-%m-%dT%H:%M:%S"
            end_date = datetime.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S")
            if (utils.date_to_timestamp(start) > utils.date_to_timestamp(end)):
                end_date = None
            return end_date
        kwargs_date = {}
        if self.schedule_start_date:
            kwargs_date.update({
                'start_date': get_start_date(self.schedule_start_date)
            })
            if self.schedule_end_date:
                end_date = get_end_date(self.schedule_start_date,
                                        self.schedule_end_date)
                kwargs_date.update({
                    'end_date': end_date
                })

        if self.schedule_date:
            return {'trigger': 'date',
                    'run_date': self.schedule_date}
        elif self.schedule_interval:
            kwargs = {'trigger': 'interval'}
            kwargs.update(kwargs_date)
            if self.schedule_interval == 'continuous':
                kwargs.update({'seconds': 1})
            else:
                val, unit = self.schedule_interval.split(' ')
                kwargs.update({unit: int(val)})
            return kwargs
        elif self.schedule_cron_fields:
            kwargs = {'trigger': 'cron'}
            kwargs.update(kwargs_date)
            cron_fields = self.schedule_cron_fields
            kwargs.update(cron_fields)
            return kwargs
        else:
            # no scheduling information, schedule to start within a few seconds
            return {'trigger': 'date',
                    'run_date': datetime.datetime.now() +
                    datetime.timedelta(0, 2, 0)}

    def process_event(self, job_doc):
        with self.scheduler.lock:
            next_event = job_doc['job_schedule'].get('event', '')
            while next_event:
                if next_event == Job.STOP_EVENT:
                    if isinstance(self.state(), StopState):
                        LOG.info('JOB {0} event: STOP'.format(self.id))
                    next_event = self.state.stop(self, job_doc)
                elif next_event == Job.START_EVENT:
                    LOG.info('JOB {0} event: START'.format(self.id))
                    next_event = self.state.start(self, job_doc)
                elif next_event == Job.ABORT_EVENT:
                    LOG.info('JOB {0} event: ABORT'.format(self.id))
                    next_event = self.state.abort(self, job_doc)
                elif next_event == Job.ABORTED_RESULT:
                    LOG.info('JOB {0} aborted.'.format(self.id))
                    break

    def upload_metadata(self, metadata_string):
        try:
            metadata = json.loads(metadata_string)
            if metadata:
                metadata['job_id'] = self.id
                self.scheduler.upload_metadata(metadata)
                LOG.info("Job {0}, freezer action metadata uploaded"
                         .format(self.id))
        except Exception as e:
            LOG.error('metrics upload error: {0}'.format(e))

    def execute_job_action(self, job_action):
        max_tries = (job_action.get('max_retries', 0) + 1)
        tries = max_tries
        freezer_action = job_action.get('freezer_action', {})
        max_retries_interval = job_action.get('max_retries_interval', 60)
        action_name = freezer_action.get('action', '')
        while tries:
            with tempfile.NamedTemporaryFile(mode='w',
                                             delete=False) as config_file:
                self.save_action_to_file(freezer_action, config_file)
                config_file_name = config_file.name
                freezer_command = '{0} --metadata-out - --config {1}'.\
                    format(self.executable, config_file.name)
                self.process = subprocess.Popen(freezer_command.split(),
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE,
                                                env=os.environ.copy())

                # store the pid for this process in the api
                try:
                    self.job_doc['job_schedule']['current_pid'] = \
                        self.process.pid

                    self.scheduler.update_job(self.job_doc['job_id'],
                                              self.job_doc)
                except Exception as error:
                    LOG.error("Error saving the process id {}".format(error))

                output, error = self.process.communicate()
                # ensure the tempfile gets deleted
                utils.delete_file(config_file_name)

            if error:
                LOG.error("Freezer client error: {0}".format(error))
            elif output:
                self.upload_metadata(output)

            if self.process.returncode == -15:
                # This means the job action was aborted by the scheduler
                LOG.warning('Freezer-agent was killed by the scheduler. '
                            'Cleanup should be done manually: container, '
                            'mountpoint and lvm snapshots.')
                return Job.ABORTED_RESULT

            elif self.process.returncode:
                # ERROR
                tries -= 1
                if tries:
                    LOG.warning('Job {0} failed {1} action,'
                                ' retrying in {2} seconds'
                                .format(self.id, action_name,
                                        max_retries_interval))
                    time.sleep(max_retries_interval)
            else:
                # SUCCESS
                LOG.info('Job {0} action {1}'
                         ' returned success exit code'.
                         format(self.id, action_name))
                return Job.SUCCESS_RESULT
        LOG.error('Job {0} action {1} failed after {2} tries'
                  .format(self.id, action_name, max_tries))

        return Job.FAIL_RESULT

    def contains_exec(self):
        jobs = self.job_doc.get('job_actions')
        for job in jobs:
            freezer_action = job.get('freezer_action')
            action = freezer_action.get('action')
            if action == 'exec':
                return True
        return False

    def update_job_schedule_doc(self, **kwargs):
        """
        Updates the job_schedule section of the job doc with the
        provided keyword args. No checks about accepted key/values
        are being made here since they may vary in the future.

        :param kwargs: keyword args to add
        :return: None
        """
        job_schedule = self.job_doc['job_schedule']
        job_schedule.update(kwargs)

    def execute(self):
        result = Job.SUCCESS_RESULT
        with self.scheduler.lock:
            LOG.info('job {0} running'.format(self.id))
            self.state = RunningState
            self.update_job_schedule_doc(status=Job.RUNNING_STATUS,
                                         result="",
                                         time_started=int(time.time()),
                                         time_ended=Job.TIME_NULL)
            self.scheduler.update_job_schedule(
                self.id,
                self.job_doc['job_schedule'])

        self.start_session()
        # if the job contains exec action and the scheduler passes the
        # parameter --disable-exec job execution should fail
        if self.contains_exec() and CONF.disable_exec:
            LOG.info("Job {0} failed because it contains exec action "
                     "and exec actions are disabled by scheduler"
                     .format(self.id))
            self.result = Job.FAIL_RESULT
            self.finish()
            return

        for job_action in self.job_doc.get('job_actions', []):
            if job_action.get('mandatory', False) or \
                    (result == Job.SUCCESS_RESULT):

                action_result = self.execute_job_action(job_action)
                if action_result in [Job.FAIL_RESULT, Job.ABORTED_RESULT]:
                    result = action_result

            else:
                freezer_action = job_action.get('freezer_action', {})
                action_name = freezer_action.get('action', '')
                LOG.warning("skipping {0} action".
                            format(action_name))
        self.result = result
        self.finish()

    def finish(self):
        self.update_job_schedule_doc(time_ended=int(time.time()))
        self.end_session(self.result)
        with self.scheduler.lock:
            if self.event == Job.REMOVE_EVENT:
                self.unschedule()
                self.job_doc_status = Job.REMOVED_STATUS
                return

            if not self.scheduled:
                self.job_doc_status = Job.COMPLETED_STATUS
                self.state = StopState
                self.scheduler.update_job(self.id, self.job_doc)
                return

            if self.event in [Job.STOP_EVENT, Job.ABORT_EVENT]:
                self.unschedule()
                self.job_doc_status = Job.COMPLETED_STATUS
                self.scheduler.update_job(self.id, self.job_doc)
            else:
                self.job_doc_status = Job.SCHEDULED_STATUS
                self.state = ScheduledState
                self.scheduler.update_job_schedule(
                    self.id,
                    self.job_doc['job_schedule'])

    def start_session(self):
        if not self.session_id:
            return
        retry = 5
        while retry:
            try:
                resp = self.scheduler.start_session(self.session_id,
                                                    self.id,
                                                    self.session_tag)
                if resp['result'] == 'success':
                    self.session_tag = resp['session_tag']
                    return
            except Exception as e:
                LOG.error('Error while starting session {0}. {1}'.
                          format(self.session_id, e))
            LOG.warning('Retrying to start session {0}'.
                        format(self.session_id))
            retry -= 1
        LOG.error('Unable to start session {0}'.format(self.session_id))

    def end_session(self, result):
        if not self.session_id:
            return
        retry = 5
        while retry:
            try:
                resp = self.scheduler.end_session(self.session_id,
                                                  self.id,
                                                  self.session_tag,
                                                  result)
                if resp['result'] == 'success':
                    return
            except Exception as e:
                LOG.error('Error while ending session {0}. {1}'.
                          format(self.session_id, e))
            LOG.warning('Retrying to end session {0}'.
                        format(self.session_id))
            retry -= 1
        LOG.error('Unable to end session {0}'.format(self.session_id))

    def schedule(self):
        try:
            kwargs = self.get_schedule_args()
            self.scheduler.add_job(self.execute, id=self.id,
                                   executor='threadpool',
                                   misfire_grace_time=3600, **kwargs)
        except Exception as e:
            LOG.error("Unable to schedule job {0}: {1}".
                      format(self.id, e))

        LOG.info('scheduler job with parameters {0}'.format(kwargs))

        if self.scheduled:
            self.job_doc_status = Job.SCHEDULED_STATUS
            self.state = ScheduledState
        else:
            # job not scheduled or already started and waiting for lock
            self.job_doc_status = Job.COMPLETED_STATUS
            self.state = StopState

    def unschedule(self):
        try:
            # already executing job are not present in the apscheduler list
            self.scheduler.remove_job(job_id=self.id)
        except Exception:
            pass
        self.event = Job.NO_EVENT
        self.job_doc_status = Job.STOP_STATUS
        self.state = StopState

    def terminate(self):
        if self.process:
            self.process.terminate()

    def kill(self):
        if self.process:
            self.process.kill()



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import os
import shutil

from freezer.storage import fslike
from freezer.utils import utils


class LocalStorage(fslike.FsLikeStorage):
    _type = 'local'

    def get_file(self, from_path, to_path):
        shutil.copyfile(from_path, to_path)

    def put_file(self, from_path, to_path):
        shutil.copyfile(from_path, to_path)

    def listdir(self, directory):
        try:
            return os.listdir(directory)
        except OSError:
            return list()

    def create_dirs(self, path):
        utils.create_dir_tree(path)

    def rmtree(self, path):
        shutil.rmtree(path)

    def open(self, filename, mode):
        return io.open(filename, mode)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from os_brick import exception
from os_brick import initiator


class VolumeAction(object):
    def __init__(self, volumes_client, volume_id):
        self.volumes_client = volumes_client
        self.volume_id = volume_id

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if traceback:
            self.volumes_client.volumes.unreserve(self.volume_id)
            return False
        return True


class Reserve(VolumeAction):
    def reserve(self):
        self.volumes_client.volumes.reserve(self.volume_id)


class InitializeConnection(VolumeAction):
    def initialize(self, brick_client, multipath, enforce_multipath):
        conn_prop = brick_client.get_connector(multipath, enforce_multipath)
        return self.volumes_client.volumes.initialize_connection(
            self.volume_id, conn_prop)


class VerifyProtocol(VolumeAction):
    # NOTE(e0ne): Only iSCSI and RBD based drivers are supported. NFS doesn't
    # work. Drivers with other protocols are not tested yet.
    SUPPORTED_PROCOTOLS = [initiator.ISCSI, initiator.RBD]

    def verify(self, protocol):
        protocol = protocol.upper()

        # NOTE(e0ne): iSCSI drivers works without issues, RBD and NFS don't
        # work. Drivers with other protocols are not tested yet.
        if protocol not in VerifyProtocol.SUPPORTED_PROCOTOLS:
            raise exception.ProtocolNotSupported(protocol=protocol)


class ConnectVolume(VolumeAction):
    def connect(self, brick_connector, connection_data,
                mountpoint, mode, hostname):
        device_info = brick_connector.connect_volume(connection_data)

        self.volumes_client.volumes.attach(self.volume_id, instance_uuid=None,
                                           mountpoint=mountpoint,
                                           mode=mode,
                                           host_name=hostname)
        return device_info


class VolumeDetachAction(VolumeAction):
    def __exit__(self, type, value, traceback):
        if traceback:
            self.volumes_client.volumes.roll_detaching(self.volume_id)
            return False
        return True


class BeginDetach(VolumeDetachAction):
    def reserve(self):
        self.volumes_client.volumes.begin_detaching(self.volume_id)


class InitializeConnectionForDetach(InitializeConnection, VolumeDetachAction):
    pass


class DisconnectVolume(VolumeDetachAction):
    def disconnect(self, brick_connector, connection_data, device_info):
        device_info = device_info or {}

        brick_connector.disconnect_volume(connection_data, device_info)


class DetachVolume(VolumeDetachAction):
    def detach(self, brick_client,
               attachment_uuid, multipath, enforce_multipath):
        conn_prop = brick_client.get_connector(multipath, enforce_multipath)

        self.volumes_client.volumes.terminate_connection(self.volume_id,
                                                         conn_prop)
        self.volumes_client.volumes.detach(self.volume_id, attachment_uuid)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer restore modes related functions
"""

import os
import shutil
import tempfile
import time

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils as json

from freezer.utils import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class RestoreOs(object):
    def __init__(self, client_manager, container, storage):
        self.client_manager = client_manager
        self.container = container
        self.storage = storage

    def _get_backups(self, path, restore_from_timestamp):
        """
        :param path:
        :type path: str
        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :return:
        """
        if self.storage.type == "swift":
            swift = self.client_manager.get_swift()
            path = "{0}_segments/{1}/".format(self.container, path)
            container_name, path = self.get_storage_info(path)
            info, backups = swift.get_container(container_name, prefix=path)
            backups = sorted(
                map(lambda x: int(x["name"].rsplit("/", 1)[-1]), backups))
        elif self.storage.type == "s3":
            bucket_name, path = self.get_storage_info(path)
            backups = self.storage.list_all_objects(
                bucket_name=bucket_name,
                prefix=path
            )
            backups = sorted(
                map(lambda x: int(x['Key'].split("/", 2)[1]), backups))
        elif self.storage.type == "local":
            path = "{0}/{1}".format(self.container, path)
            backups = os.listdir(os.path.abspath(path))
        elif self.storage.type in ['ssh', 'ftp', 'ftps']:
            path = "{0}/{1}".format(self.container, path)
            backups = self.storage.listdir(path)
        else:
            msg = ("{} storage type is not supported at the moment."
                   " Try local, SWIFT, SSH(SFTP), FTP or FTPS ".
                   format(self.storage.type))
            LOG.error(msg)
            raise BaseException(msg)
        backups = list(filter(lambda x: int(x) >= restore_from_timestamp,
                              backups))
        if not backups:
            msg = "Cannot find backups for path: %s" % path
            LOG.error(msg)
            raise BaseException(msg)
        return backups[-1]

    def get_storage_info(self, path):
        storage_info = self.container.split('/', 1)
        container_name = storage_info[0]
        object_prefix = storage_info[1] if '/' in self.container else ''
        if object_prefix != '':
            path = "{0}/{1}".format(object_prefix, path)
        return container_name, path

    def _create_image(self, path, restore_from_timestamp):
        """
        :param path:
        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :return:
        """
        backup = self._get_backups(path, restore_from_timestamp)
        if self.storage.type == 'swift':
            swift = self.client_manager.get_swift()
            path = "{0}_segments/{1}/{2}".format(self.container, path, backup)
            stream = swift.get_object(self.container,
                                      "{}/{}".format(path, backup),
                                      resp_chunk_size=10000000)
            length = int(stream[0]["x-object-meta-length"])
            data = utils.ReSizeStream(stream[1], length, 1)
            info = stream[0]
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 's3':
            if self.storage.get_object_prefix() != '':
                base_path = "{0}/{1}/{2}".format(
                    self.storage.get_object_prefix(),
                    path,
                    backup
                )
            else:
                base_path = "{0}/{1}".format(path, backup)
            image_file = "{0}/{1}".format(base_path, path)
            s3_object = self.storage.get_object(
                bucket_name=self.storage.get_bucket_name(),
                key=image_file
            )
            stream = utils.S3ResponseStream(data=s3_object['Body'],
                                            chunk_size=10000000)
            data = utils.ReSizeStream(
                stream,
                s3_object['ContentLength'],
                1
            )
            metadata = "{0}/metadata".format(base_path)
            metadata_object = self.storage.get_object(
                bucket_name=self.storage.get_bucket_name(),
                key=metadata
            )
            info = json.loads(metadata_object['Body'])

            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 'local':
            image_file = "{0}/{1}/{2}/{3}".format(self.container, path,
                                                  backup, path)
            metadata_file = "{0}/{1}/{2}/metadata".format(self.container,
                                                          path, backup)
            try:
                data = open(image_file, 'rb')
            except Exception:
                msg = "Failed to open image file {}".format(image_file)
                LOG.error(msg)
                raise BaseException(msg)
            info = json.load(open(metadata_file, 'r'))
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 'ssh':
            image_file = "{0}/{1}/{2}/{3}".format(self.container, path,
                                                  backup, path)
            metadata_file = "{0}/{1}/{2}/metadata".format(self.container,
                                                          path, backup)
            try:
                data = self.storage.open(image_file, 'rb')
            except Exception:
                msg = "Failed to open remote image file {}".format(image_file)
                LOG.error(msg)
                raise BaseException(msg)
            info = json.loads(self.storage.read_metadata_file(metadata_file))
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type in ['ftp', 'ftps']:
            image_file = "{0}/{1}/{2}/{3}".format(self.container, path,
                                                  backup, path)
            metadata_file = "{0}/{1}/{2}/metadata".format(self.container,
                                                          path, backup)
            try:
                tmpdir = tempfile.mkdtemp()
            except Exception:
                LOG.error("Unable to create a tmp directory")
                raise Exception("Unable to create a tmp directory")
            try:
                data_image = utils.path_join(tmpdir, "data_image")
                LOG.info('create image restore ftp storage')
                self.storage.get_file(image_file, data_image)
                data_meta = utils.path_join(tmpdir, "data_meta")
                self.storage.get_file(metadata_file, data_meta)
                data = open(data_image, 'rb')
                info = json.load(open(data_meta, 'r'))
                image = self.client_manager.create_image(
                    name="restore_{}".format(path),
                    container_format="bare",
                    disk_format="raw",
                    data=data)
                return info, image
            finally:
                shutil.rmtree(tmpdir)
        else:
            return {}

    def restore_cinder(self, volume_id=None,
                       backup_id=None,
                       restore_from_timestamp=None):
        """
        Restoring cinder backup using
        :param volume_id:
        :param backup_id:
        :param restore_from_timestamp:
        :return:
        """
        cinder = self.client_manager.get_cinder()
        search_opts = {
            'volume_id': volume_id,
            'status': 'available',
        }
        if not backup_id:
            backups = cinder.backups.list(search_opts=search_opts)

            def get_backups_from_timestamp(backups, restore_from_timestamp):
                for backup in backups:
                    backup_created_date = backup.created_at.split('.')[0]
                    backup_created_timestamp = (
                        utils.date_to_timestamp(backup_created_date))
                    if backup_created_timestamp >= restore_from_timestamp:
                        yield backup

            backups_filter = get_backups_from_timestamp(backups,
                                                        restore_from_timestamp)
            if not backups_filter:
                LOG.warning("no available backups for cinder volume,"
                            "restore newest backup")
                backup = max(backups, key=lambda x: x.created_at)
            else:
                backup = min(backups_filter, key=lambda x: x.created_at)
            backup_id = backup.id
        cinder.restores.restore(backup_id=backup_id)

    def restore_cinder_by_glance(self, volume_id, restore_from_timestamp):
        """
        1) Define swift directory
        2) Download and upload to glance
        3) Create volume from glance
        4) Delete

        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :param volume_id: - id of attached cinder volume
        """
        (info, image) = self._create_image(volume_id, restore_from_timestamp)
        length = int(info["x-object-meta-length"])
        gb = 1073741824
        size = length / gb
        if length % gb > 0:
            size += 1
        LOG.info("Creating volume from image")
        cinder_client = self.client_manager.get_cinder()
        volume = cinder_client.volumes.create(
            size,
            imageRef=image.id,
            name=info.get('volume_name',
                          CONF.get('backup_name',
                                   CONF.get('cinder_vol_id', None)
                                   )
                          )
        )
        while volume.status != "available":
            try:
                LOG.info("Volume copy status: " + volume.status)
                volume = cinder_client.volumes.get(volume.id)
                if volume.status == "error":
                    raise Exception("Volume copy status: error")
                time.sleep(5)
            except Exception as e:
                LOG.exception(e)
                if volume.status != "error":
                    LOG.warning("Exception getting volume status")

        LOG.info("Deleting temporary image {}".format(image.id))
        self.client_manager.get_glance().images.delete(image.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file is execfile()d with the current directory set
# to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

# NOTE(dims): monkey patch subprocess to prevent failures in latest eventlet
# See https://github.com/eventlet/eventlet/issues/398


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('../../'))
# sys.path.insert(0, os.path.abspath('../'))
# sys.path.insert(0, os.path.abspath('./'))

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings.
# They can be extensions coming with Sphinx (named 'sphinx.ext.*')
# or your custom ones.

extensions = ['sphinxcontrib.apidoc',
              'openstackdocstheme',
              'oslo_config.sphinxconfiggen',
              'oslo_config.sphinxext'
              ]

config_generator_config_file = (
    '../../etc/config-generator.conf')
sample_config_basename = '_static/freezer'


# autodoc generation is a bit aggressive and a nuisance
# when doing heavy text edit cycles. Execute "export SPHINX_DEBUG=1"
# in your terminal to disable
#if not os.getenv('SPHINX_DEBUG'):
#    extensions += ['ext.freezer_autodoc']

todo_include_todos = True

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../freezer'
apidoc_output_dir = 'reference/api'
apidoc_excluded_paths = [
    'scheduler/win*']
apidoc_separate_modules = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = []

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
openstackdocs_auto_name = False
openstackdocs_repo_name = 'openstack/freezer'
openstackdocs_bug_project = 'freezer'
openstackdocs_bug_tag = 'doc'
project = u'Freezer'
copyright = u'OpenStack Foundation'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# from freezer import version_info
# # The full version, including alpha/beta/rc tags.
# release = version_info.release_string()
# # The short X.Y version.
# version = version_info.version_string()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.

# # List of directories, relative to source directory, that shouldn't be searched
# # for source files.
# exclude_trees = []

# A list of glob-style patterns that should be excluded when looking for
# source files. They are matched against the source file names relative to the
# source directory, using slashes as directory separators on all platforms.
exclude_patterns = [
    # Missing win32serviceutil module on linux
    #'api/freezer.scheduler.win_daemon*',
]

#
# # The reST default role (used for this markup: `text`) to use
# # for all documents.
# # default_role = None
#
# # If true, '()' will be appended to :func: etc. cross-reference text.
# # add_function_parentheses = True
#
# # If true, the current module name will be prepended to all description
# # unit titles (such as .. function::).
# add_module_names = False
#
# # If true, sectionauthor and moduleauthor directives will be shown in the
# # output. They are ignored by default.
# show_authors = False
#
# # The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['freezer.']

# -- Options for man page output ----------------------------------------------

# Grouping the document tree for man pages.
# List of tuples 'sourcefile', 'target', u'title', u'Authors name', 'manual'

man_pages = [
    ('man/freezer-manage', 'freezer-manage', u'Cloud controller fabric',
     [u'OpenStack'], 1)
]

# -- Options for HTML output --------------------------------------------------

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
html_theme = 'openstackdocs'



# (c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from freezer.openstack import osclients
from freezer.utils import config
from oslo_config import cfg
from oslo_log import log

CONF = cfg.CONF
LOG = log.getLogger(__name__)


def parse_osrc(file_name):
    with open(file_name, 'r') as osrc_file:
        return config.osrc_parse(osrc_file.read())


def get_client_manager(backup_args):
    if "osrc" in backup_args:
        options = osclients.OpenstackOpts.create_from_dict(
            parse_osrc(backup_args['osrc']))
    else:
        options = osclients.OpenstackOpts.create_from_env().get_opts_dicts()
    if backup_args['project_id']:
        options['project_name'] = None
        options['project_id'] = backup_args['project_id']
    client_manager = osclients.OSClientManager(
        auth_url=options.pop('auth_url', None),
        auth_method=options.pop('auth_method', 'password'),
        dry_run=backup_args.get('dry_run', None),
        **options
    )
    return client_manager



# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import signal
import traceback

from oslo_log import log
from tempfile import gettempdir
from time import sleep


from freezer.lib.pep3143daemon import DaemonContext
from freezer.lib.pep3143daemon import PidFile


LOG = log.getLogger(__name__)


def get_filenos(logger):
    """
    Get a list of file no from logger
    """
    filenos = []
    for handler in logger.handlers:
        filenos.append(handler.stream.fileno())
    if logger.parent:
        filenos += get_filenos(logger.parent)
    return filenos


def is_process_running(pid):
    """
    Checks whether the process is running.

    :param pid: process pid to check
    :return: true if the process is running
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


class NoDaemon(object):
    """
    A class which shares the same interface as the Daemon class,
    but is used to execute the scheduler as a foreground process

    """

    instance = None
    exit_flag = False

    def __init__(self, daemonizable):
        # daemonizable has to provide start/stop (and possibly reload) methods
        NoDaemon.instance = self
        self.daemonizable = daemonizable

        # register signal handlers
        for (signal_number, handler) in self.signal_map.items():
            signal.signal(signal_number, handler)

    @property
    def signal_map(self):
        return {
            signal.SIGTERM: NoDaemon.handle_program_exit,
            signal.SIGINT: NoDaemon.handle_program_exit,
            signal.SIGHUP: NoDaemon.handle_reload,
        }

    @staticmethod
    def handle_program_exit(signum, frame):
        LOG.info('Got signal {0}. Exiting ...'.format(signum))
        NoDaemon.exit_flag = True
        NoDaemon.instance.daemonizable.stop()

    @staticmethod
    def handle_reload(signum, frame):
        NoDaemon.instance.daemonizable.reload()

    def start(self, dump_stack_trace=False):
        while not NoDaemon.exit_flag:
            try:
                LOG.info('Starting in no-daemon mode')
                self.daemonizable.start()
                NoDaemon.exit_flag = True
            except Exception as e:
                if dump_stack_trace:
                    LOG.error(traceback.format_exc(e))
                LOG.error('Restarting procedure in no-daemon mode '
                          'after Fatal Error: {0}'.format(e))
                sleep(10)
        LOG.info('Done exiting')

    def stop(self):
        pass

    def status(self):
        pass

    def restart(self):
        pass

    def reload(self):
        pass


class Daemon(object):
    """
    A class to manage all the daemon-related stuff

    """

    instance = None
    exit_flag = False

    def __init__(self, daemonizable=None, pid_fname=None):
        # daemonizable has to provide start/stop (and possibly reload) methods
        Daemon.instance = self
        self._pid_fname = pid_fname
        self.daemonizable = daemonizable

    @staticmethod
    def handle_program_exit(signum, frame):
        Daemon.exit_flag = True
        Daemon.instance.daemonizable.stop()

    @staticmethod
    def handle_reload(signum, frame):
        Daemon.instance.daemonizable.reload()

    @property
    def signal_map(self):
        return {
            signal.SIGTERM: Daemon.handle_program_exit,
            signal.SIGHUP: Daemon.handle_reload,
        }

    @property
    def pid_fname(self):
        if not self._pid_fname:
            fname = '{0}/freezer_sched_{1}.pid'.format(
                gettempdir(),
                os.path.split(os.path.expanduser('~'))[-1])
            self._pid_fname = os.path.normpath(fname)
        return self._pid_fname

    @property
    def pid(self):
        if os.path.isfile(self.pid_fname):
            with open(self.pid_fname, 'r') as f:
                return int(f.read())
        return None

    def start(self, dump_stack_trace=False):
        if os.path.exists(self.pid_fname) and \
                is_process_running(self.pid):
            print('freezer daemon is already running, '
                  'pid: {0}'.format(self.pid))
            return
        pidfile = PidFile(self.pid_fname)
        files_preserve = get_filenos(LOG.logger)
        with DaemonContext(pidfile=pidfile, signal_map=self.signal_map,
                           files_preserve=files_preserve):
            while not Daemon.exit_flag:
                try:
                    LOG.info('freezer daemon starting, pid: {0}'.
                             format(self.pid))
                    self.daemonizable.start()
                    Daemon.exit_flag = True
                except Exception as e:
                    if dump_stack_trace:
                        LOG.error(traceback.format_exc(e))
                    LOG.error('Restarting daemonized procedure '
                              'after Fatal Error: {0}'.format(e))
                    sleep(10)
            LOG.info('freezer daemon done, pid: {0}'.format(self.pid))

    def stop(self):
        pid = self.pid
        if pid:
            os.kill(self.pid, signal.SIGTERM)
        else:
            print('Not Running')

    def restart(self):
        pid = self.pid
        if not pid:
            self.start()
        else:
            self.stop()
            sleep(5)
            self.start()

    def status(self):
        pid = self.pid
        if pid:
            print('Running with pid: {0}'.format(pid))
        else:
            print('Not Running')

    def reload(self):
        pid = self.pid
        if pid:
            os.kill(pid, signal.SIGHUP)
        else:
            print('Not Running')



# (C) Copyright 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes
from os import urandom

SALT_HEADER = 'Salted__'
AES256_KEY_LENGTH = 32
BS = 16  # static 16 bytes, 128 bits for AES


class AESCipher(object):
    """
    Base class for encrypt/decrypt activities.
    """

    def __init__(self, pass_file):
        self._password = self._get_pass_from_file(pass_file)
        self._salt = None
        self.cipher = None

    @staticmethod
    def _get_pass_from_file(pass_file):
        with open(pass_file) as p_file:
            password = p_file.readline()
        return password

    @staticmethod
    def _derive_key_and_iv(password, salt, key_length, iv_length):
        d = d_i = b''
        while len(d) < key_length + iv_length:
            md5_str = d_i + password + salt
            d_i = hashlib.md5(md5_str).digest()
            d += d_i
        return d[:key_length], d[key_length:key_length + iv_length]


class AESEncrypt(AESCipher):
    """
    Encrypts chucks of data using AES-256 algorithm.
    OpenSSL compatible.
    """

    def __init__(self, pass_file):
        super(AESEncrypt, self).__init__(pass_file)
        self._salt = urandom(BS - len(SALT_HEADER))
        key, iv = self._derive_key_and_iv(self._password,
                                          self._salt,
                                          AES256_KEY_LENGTH,
                                          BS)
        self.cipher = Cipher(algorithms.AES(key),
                             modes.CFB(iv),
                             backend=default_backend())

    def generate_header(self):
        return SALT_HEADER + self._salt

    def encrypt(self, data):
        encryptor = self.cipher.encryptor()
        return encryptor.update(data) + encryptor.finalize()


class AESDecrypt(AESCipher):
    """
    Decrypts chucks of data using AES-256 algorithm.
    OpenSSL compatible.
    """

    def __init__(self, pass_file, salt):
        super(AESDecrypt, self).__init__(pass_file)
        self._salt = self._prepare_salt(salt)
        key, iv = self._derive_key_and_iv(self._password,
                                          self._salt,
                                          AES256_KEY_LENGTH,
                                          BS)
        self.cipher = Cipher(algorithms.AES(key),
                             modes.CFB(iv),
                             backend=default_backend())

    @staticmethod
    def _prepare_salt(salt):
        return salt[len(SALT_HEADER):]

    def decrypt(self, data):
        decryptor = self.cipher.decryptor()
        return decryptor.update(data) + decryptor.finalize()



# (c) Copyright 2018 ZTE Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import tempfile
import unittest
from unittest import mock
from unittest.mock import patch

from freezer.storage import ftp


class BaseFtpStorageTestCase(unittest.TestCase):

    def setUp(self):
        super(BaseFtpStorageTestCase, self).setUp()
        self.ftp_opt = mock.Mock()
        self.ftp_opt.ftp_storage_path = '/just/a/path'
        self.ftp_opt.ftp_remote_pwd = 'passawd'
        self.ftp_opt.ftp_remote_username = 'usrname'
        self.ftp_opt.ftp_remote_ip = '0.0.0.0'
        self.ftp_opt.ftp_port = 2121
        self.ftp_opt.ftp_max_segment_size = 1024

    def test_validate_BaseFtpStorage(self):
        with self.assertRaises(Exception) as cm:  # noqa
            ftp.BaseFtpStorage(
                storage_path=self.ftp_opt.ftp_storage_path,
                remote_pwd=self.ftp_opt.ftp_remote_pwd,
                remote_ip=None,
                remote_username=self.ftp_opt.ftp_remote_username,
                port=self.ftp_opt.ftp_port,
                max_segment_size=self.ftp_opt.ftp_max_segment_size)
            the_exception = cm.exception
            self.assertIn('Please provide --ftp-host value',
                          str(the_exception))

        with self.assertRaises(Exception) as cm:  # noqa
            ftp.BaseFtpStorage(
                storage_path=self.ftp_opt.ftp_storage_path,
                remote_pwd=self.ftp_opt.ftp_remote_pwd,
                remote_ip=self.ftp_opt.ftp_remote_ip,
                remote_username=None,
                port=self.ftp_opt.ftp_port,
                max_segment_size=self.ftp_opt.ftp_max_segment_size)
            the_exception = cm.exception
            self.assertIn('Please provide --ftp-username value',
                          str(the_exception))

        with self.assertRaises(Exception) as cm:  # noqa
            ftp.BaseFtpStorage(
                storage_path=self.ftp_opt.ftp_storage_path,
                remote_pwd=None,
                remote_username=self.ftp_opt.ftp_remote_username,
                remote_ip=self.ftp_opt.ftp_remote_ip,
                port=self.ftp_opt.ftp_port,
                max_segment_size=self.ftp_opt.ftp_max_segment_size)
            the_exception = cm.exception
            self.assertIn('Please provide remote password',
                          str(the_exception))


class FtpStorageTestCase(unittest.TestCase):

    def setUp(self):
        super(FtpStorageTestCase, self).setUp()
        self.ftp_opt = mock.Mock()
        self.ftp_opt.ftp_storage_path = '/just/a/path'
        self.ftp_opt.ftp_remote_pwd = 'passawd'
        self.ftp_opt.ftp_remote_username = 'usrname'
        self.ftp_opt.ftp_remote_ip = '0.0.0.0'
        self.ftp_opt.ftp_port = 2121
        self.ftp_test_file_dir = None
        self.ftp_test_file_name = None

    def create_ftp_test_file(self):
        if self.ftp_test_file_name:
            return
        tmpdir = tempfile.mkdtemp()
        FILES_DIR_PREFIX = "freezer_ftptest_files_dir"
        files_dir = tempfile.mkdtemp(dir=tmpdir, prefix=FILES_DIR_PREFIX)
        file_name = "file_ftp_test"
        self.ftp_test_file_dir = files_dir
        text = "FTPTESTTXT"
        filehandle = open(files_dir + "/" + file_name, 'w')
        if filehandle:
            filehandle.write(text)
            filehandle.close()
            self.ftp_test_file_name = file_name

    def delete_ftp_test_file(self):
        if self.ftp_test_file_name:
            files_dir = self.ftp_test_file_dir
            shutil.rmtree(files_dir)
            self.ftp_test_file_name = None
            self.ftp_test_file_dir = None

    def create_ftpstorage_obj(self):
        obj = ftp.FtpStorage(
            storage_path=self.ftp_opt.ftp_storage_path,
            remote_pwd=self.ftp_opt.ftp_remote_pwd,
            remote_username=self.ftp_opt.ftp_remote_username,
            remote_ip=self.ftp_opt.ftp_remote_ip,
            port=self.ftp_opt.ftp_port,
            max_segment_size=self.ftp_opt.ftp_max_segment_size)
        return obj

    @patch('ftplib.FTP')
    def test_init_fail_raise_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        seffect = mock.Mock(side_effect=Exception('create ftp failed error'))
        mock_ftp.raiseError.side_effect = seffect
        with self.assertRaises(Exception) as cm:  # noqa
            self.create_ftpstorage_obj()
            the_exception = cm.exception
            self.assertIn('create ftp failed error',
                          str(the_exception))

    @patch('ftplib.FTP')
    def test_init_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        self.create_ftpstorage_obj()
        self.assertTrue(mock_ftp.set_pasv.called)
        self.assertTrue(mock_ftp.connect.called)
        self.assertTrue(mock_ftp.login.called)
        self.assertTrue(mock_ftp.nlst.called)
        mock_ftp.set_pasv.assert_called_with(True)
        mock_ftp.connect.assert_called_with(self.ftp_opt.ftp_remote_ip,
                                            self.ftp_opt.ftp_port, 60)
        mock_ftp.login.assert_called_with(self.ftp_opt.ftp_remote_username,
                                          self.ftp_opt.ftp_remote_pwd)

    @patch('tempfile.mkdtemp')
    @patch('ftplib.FTP')
    def test_create_tempdir_FtpStorage(self, mock_ftp_constructor,
                                       mock_tempfile_constructor):
        mock_tempfile = mock_tempfile_constructor.return_value
        ftp_obj = self.create_ftpstorage_obj()
        tmp = ftp_obj._create_tempdir()
        self.assertEqual(tmp, mock_tempfile)

    @patch('ftplib.FTP')
    def test_listdir_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        test_dir = '/home/test'
        filelist = ['test1.py', 'test2.py', 'readme.txt', 'ftp.sh']
        mock_ftp.nlst.return_value = filelist
        ret = ftpobj.listdir(test_dir)
        self.assertTrue(mock_ftp.nlst.called)
        mock_ftp.cwd.assert_called_with(test_dir)
        self.assertEqual(sorted(filelist), ret)

    @patch('ftplib.FTP')
    def test_listdir_fail_raise_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        test_dir = '/home/test'
        seffect = mock.Mock(side_effect=Exception())
        mock_ftp.raiseError.side_effect = seffect
        ret = ftpobj.listdir(test_dir)
        self.assertEqual(list(), ret)

    @patch('ftplib.FTP')
    def test_putfile_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        self.create_ftp_test_file()
        frompath = self.ftp_test_file_dir + "/" + self.ftp_test_file_name
        topath = '/home/to'
        ftpobj.put_file(frompath, topath)
        self.assertTrue(mock_ftp.pwd.called)
        self.assertTrue(mock_ftp.storbinary.called)
        self.delete_ftp_test_file()

    @patch('ftplib.FTP')
    def test_getfile_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        self.create_ftp_test_file()
        topath = self.ftp_test_file_dir + "/" + self.ftp_test_file_name
        frompath = '/home/from'
        ftpobj.get_file(frompath, topath)
        self.assertTrue(mock_ftp.pwd.called)
        self.assertTrue(mock_ftp.retrbinary.called)
        self.delete_ftp_test_file()

    @patch('ftplib.FTP')
    def test_create_dirs_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        path = '/'
        ftpobj.create_dirs(path)
        mock_ftp.cwd.assert_called_with(path)
        path = '/home'
        ftpobj.create_dirs(path)
        mock_ftp.cwd.assert_called_with(path)

    @patch('ftplib.FTP')
    def test_rmtree_ok_FtpStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        ftpobj = self.create_ftpstorage_obj()
        path = '/home/tecs'
        ftpobj.rmtree(path)
        self.assertTrue(mock_ftp.dir.called)
        self.assertTrue(mock_ftp.rmd.called)

    @patch("freezer.storage.ftp.BaseFtpStorage.create_dirs")
    @patch("freezer.storage.ftp.BaseFtpStorage.put_file")
    @patch("builtins.open")
    @patch('ftplib.FTP')
    def test_write_backup_FtpStorage(self, mock_ftp_constructor,
                                     mock_open,
                                     mock_put_file,
                                     mock_create_dirs):
        ftpobj = self.create_ftpstorage_obj()

        rich_queue = mock.MagicMock()
        rich_queue.get_messages = mock.MagicMock()
        rich_queue.get_messages.return_value = ['1']

        backup = mock.MagicMock()
        backup.copy = mock.MagicMock()
        backup.copy.return_value = backup

        path = 'fakepath'
        backup.data_path = mock.MagicMock()
        backup.data_path.return_value = path

        b_file = mock.MagicMock()
        b_file.write = mock.MagicMock()

        mock_open = mock.MagicMock()
        mock_open.return_value = b_file

        ftpobj.write_backup(rich_queue=rich_queue,
                            backup=backup)
        self.assertTrue(mock_create_dirs.called)
        self.assertTrue(mock_put_file.called)

    @patch("freezer.storage.ftp.BaseFtpStorage.create_dirs")
    @patch("freezer.storage.ftp.BaseFtpStorage.put_file")
    @patch("builtins.open")
    @patch('ftplib.FTP')
    def test_add_stream_FtpStorage(self, mock_ftp_constructor,
                                   mock_open,
                                   mock_put_file,
                                   mock_create_dirs):
        ftpobj = self.create_ftpstorage_obj()

        rich_queue = mock.MagicMock()
        rich_queue.get_messages = mock.MagicMock()
        rich_queue.get_messages.return_value = ['1']

        package_name = 'fakedir/fakename'
        stream = ['fakestream']

        backup = mock.MagicMock()
        backup.copy = mock.MagicMock()
        backup.copy.return_value = backup

        b_file = mock.MagicMock()
        b_file.write = mock.MagicMock()

        mock_open = mock.MagicMock()
        mock_open.return_value = b_file

        ftpobj.add_stream(stream=stream,
                          package_name=package_name)

        self.assertTrue(mock_create_dirs.called)
        self.assertTrue(mock_put_file.called)


class FtpsStorageTestCase(unittest.TestCase):

    def setUp(self):
        super(FtpsStorageTestCase, self).setUp()
        self.ftp_opt = mock.Mock()
        self.ftp_opt.ftp_storage_path = '/just/a/path'
        self.ftp_opt.ftp_remote_pwd = 'passawd'
        self.ftp_opt.ftp_remote_username = 'usrname'
        self.ftp_opt.ftp_remote_ip = '0.0.0.0'
        self.ftp_opt.ftp_port = 2121
        self.ftp_opt.ftp_keyfile = '/just/key.pem'
        self.ftp_opt.ftp_certfile = '/just/cert.pem'

    def create_ftpsStorage_obj(self):
        obj = ftp.FtpsStorage(
            storage_path=self.ftp_opt.ftp_storage_path,
            remote_pwd=self.ftp_opt.ftp_remote_pwd,
            remote_username=self.ftp_opt.ftp_remote_username,
            remote_ip=self.ftp_opt.ftp_remote_ip,
            port=self.ftp_opt.ftp_port,
            max_segment_size=self.ftp_opt.ftp_max_segment_size,
            keyfile=self.ftp_opt.ftp_keyfile,
            certfile=self.ftp_opt.ftp_certfile)
        return obj

    def test_init_fail_FtpsStorage(self):
        with self.assertRaises(Exception) as cm:  # noqa
            self.create_ftpsStorage_obj()
            the_exception = cm.exception
            self.assertIn('create ftp failed error',
                          str(the_exception))

    @patch('ftplib.FTP_TLS')
    def test_init_ok_FtpsStorage(self, mock_ftp_constructor):
        mock_ftp = mock_ftp_constructor.return_value
        self.create_ftpsStorage_obj()
        self.assertTrue(mock_ftp.set_pasv.called)
        self.assertTrue(mock_ftp.connect.called)
        self.assertTrue(mock_ftp.login.called)
        self.assertTrue(mock_ftp.prot_p.called)
        self.assertTrue(mock_ftp.nlst.called)
        mock_ftp.set_pasv.assert_called_with(True)
        mock_ftp.connect.assert_called_with(self.ftp_opt.ftp_remote_ip,
                                            self.ftp_opt.ftp_port, 60)
        mock_ftp.login.assert_called_with(self.ftp_opt.ftp_remote_username,
                                          self.ftp_opt.ftp_remote_pwd)



# (c) Copyright 2019 ZTE Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import unittest
from unittest import mock
from unittest.mock import patch

from freezer.scheduler import utils

job_list = [{"job_id": "test"}]


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.client.clients.create = mock.Mock(return_value='test')
        self.client.jobs.list = mock.Mock(return_value=job_list)
        self.client.client_id = "test"

    def test_do_register(self):
        ret = utils.do_register(self.client, args=None)
        self.assertEqual(0, ret)

    def test_del_register_error(self):
        self.client.clients.delete = mock.Mock(side_effect=Exception(
            'delete client error: bad request'))
        with self.assertRaises(Exception) as cm:  # noqa
            utils.del_register(self.client)
            the_exception = cm.exception
            self.assertIn('delete client error', str(the_exception))

    def test_find_config_files(self):
        temp = tempfile.NamedTemporaryFile('wb', delete=True,
                                           suffix='.conf')
        ret = utils.find_config_files(temp.name)
        self.assertEqual([temp.name], ret)
        temp.close()
        self.assertFalse(os.path.exists(temp.name))

    def test_find_config_files_path(self):
        temp = tempfile.NamedTemporaryFile('wb', delete=True,
                                           suffix='.conf')
        temp_path = os.path.dirname(temp.name)
        ret = utils.find_config_files(temp_path)
        self.assertEqual([temp.name], ret)
        temp.close()
        self.assertFalse(os.path.exists(temp.name))

    @patch('oslo_serialization.jsonutils.load')
    def test_load_doc_from_json_file(self, mock_load):
        os.mknod("/tmp/test_freezer.conf")
        mock_load.side_effect = Exception('error')
        try:
            utils.load_doc_from_json_file("/tmp/test_freezer.conf")
        except Exception as e:
            self.assertIn("Unable to load conf file", str(e))
        os.remove("/tmp/test_freezer.conf")

    def test_get_jobs_from_disk(self):
        temp = tempfile.mkdtemp()
        file = '/'.join([temp, "test.conf"])
        data = b'{"job_id": "test"}'
        with open(file, 'wb') as f:
            f.write(data)
        ret = utils.get_jobs_from_disk(temp)
        self.assertEqual(job_list, ret)
        shutil.rmtree(temp)
        self.assertFalse(os.path.exists(file))

    def test_save_jobs_to_disk(self):
        job_doc_list = job_list
        tmpdir = tempfile.mkdtemp()
        utils.save_jobs_to_disk(job_doc_list, tmpdir)
        file = '/'.join([tmpdir, "job_test.conf"])
        self.assertTrue(os.path.exists(file))
        shutil.rmtree(tmpdir)

    def test_get_active_jobs_from_api(self):
        ret = utils.get_active_jobs_from_api(self.client)
        self.assertEqual(job_list, ret)

    @patch('os.kill')
    @patch('psutil.Process')
    def test_terminate_subprocess1(self, mock_process, mock_oskill):
        mock_pro = mock.MagicMock()
        mock_pro.name.startswith.return_value = False
        mock_process.return_value = mock_pro
        result = utils.terminate_subprocess(35, 'test')
        self.assertIsNone(result)
        mock_pro.name.startswith.return_value = True
        mock_oskill.side_effect = Exception("error")
        result = utils.terminate_subprocess(35, 'test')
        self.assertIsNone(result)

    @patch('psutil.Process')
    def test_terminate_subprocess(self, mock_process_constructor):
        mock_pro = mock_process_constructor.return_value
        seffect = mock.Mock(
            side_effect=Exception('Process 35 does not exists anymore'))
        mock_pro.raiseError.side_effect = seffect
        with self.assertRaises(Exception) as cm:  # noqa
            utils.terminate_subprocess(35, "test")
            the_exception = cm.exception
            self.assertIn('does not exists anymore',
                          str(the_exception))



class TimeoutException(Exception):
    msg = "Timeout has been occurred."

    def __init__(self, message=None, **kwargs):
        if not message:
            message = self.msg
        super(TimeoutException, self).__init__(message, kwargs)



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import unittest

from freezer.utils import config


class TestConfig(unittest.TestCase):
    def test_export(self):
        string = """unset OS_DOMAIN_NAME
export OS_AUTH_URL="http://abracadabra/v3"
export OS_PROJECT_NAME=abracadabra_project
export OS_USERNAME=abracadabra_username
export OS_PASSWORD=abracadabra_password
export OS_PROJECT_DOMAIN_NAME=Default
export OS_USER_DOMAIN_NAME=Default
export OS_IDENTITY_API_VERSION=3
export OS_AUTH_VERSION=3
export OS_CACERT=/etc/ssl/certs/ca-certificates.crt
export OS_ENDPOINT_TYPE=internalURL"""

        res = config.osrc_parse(string)
        self.assertEqual("http://abracadabra/v3", res["OS_AUTH_URL"])

    def test_ini(self):
        string = """[default]
# This is a comment line
#
host = 127.0.0.1
port = 3306
user = openstack
password = 'aNiceQuotedPassword'
password2 = "aNiceQuotedPassword"
spaced =   value"""

        fd = io.StringIO(string)
        res = config.ini_parse(fd)
        self.assertEqual('127.0.0.1', res['host'])
        self.assertEqual('openstack', res['user'])
        self.assertEqual('3306', res['port'])

        # python 3.4 tests will fail because aNiceQuatedPassword will
        # be quoted like "'aNiceQuotedPassword'" and '"aNiceQuotedPassword"'.
        # Solution for now is to strip the inside quotation marks.
        self.assertEqual('aNiceQuotedPassword', res['password'].strip("\"")
                         .strip('\''))
        self.assertEqual('aNiceQuotedPassword', res['password2'].strip("\"")
                         .strip('\''))

        self.assertEqual('value', res['spaced'])



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Freezer Backup modes related functions
"""

from oslo_config import cfg
from oslo_log import log

from freezer.utils import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class BackupOs(object):

    def __init__(self, client_manager, container, storage):
        """

        :param client_manager:
        :param container:
        :param storage:
        :type storage: freezer.swift.SwiftStorage
        :return:
        """
        self.client_manager = client_manager
        self.container = container
        self.storage = storage

    def backup_cinder_by_glance(self, volume_id):
        """
        Implements cinder backup:
            1) Gets a stream of the image from glance
            2) Stores resulted image to the swift as multipart object

        :param volume_id: id of volume for backup
        """
        client_manager = self.client_manager
        cinder = client_manager.get_cinder()

        volume = cinder.volumes.get(volume_id)
        LOG.debug("Creation temporary snapshot")
        snapshot = client_manager.provide_snapshot(
            volume, "backup_snapshot_for_volume_%s" % volume_id)
        LOG.debug("Creation temporary volume")
        copied_volume = client_manager.do_copy_volume(snapshot)
        LOG.debug("Creation temporary glance image")
        image = client_manager.make_glance_image(copied_volume.id,
                                                 copied_volume)
        LOG.debug("Download temporary glance image {0}".format(image.id))
        stream = client_manager.download_image(image)
        package = "{0}/{1}".format(volume_id, utils.DateTime.now().timestamp)
        LOG.debug("Saving image to {0}".format(self.storage.type))
        if volume.name is None:
            name = volume_id
        else:
            name = volume.name
        headers = {'x-object-meta-length': str(len(stream)),
                   'volume_name': name,
                   'availability_zone': volume.availability_zone
                   }
        attachments = volume._info['attachments']
        if attachments:
            headers['server'] = attachments[0]['server_id']
        self.storage.add_stream(stream, package, headers=headers)
        LOG.debug("Deleting temporary snapshot")
        client_manager.clean_snapshot(snapshot)
        LOG.debug("Deleting temporary volume")
        cinder.volumes.delete(copied_volume)
        LOG.debug("Deleting temporary image")
        client_manager.get_glance().images.delete(image.id)

    def backup_cinder(self, volume_id, name=None, description=None,
                      incremental=False):
        client_manager = self.client_manager
        cinder = client_manager.get_cinder()
        container = "{0}/{1}/{2}".format(self.container, volume_id,
                                         utils.DateTime.now().timestamp)
        if incremental:
            search_opts = {
                'volume_id': volume_id,
                'status': 'available'
            }
            backups = cinder.backups.list(search_opts=search_opts)
            if len(backups) <= 0:
                cinder.backups.create(volume_id, container, name, description,
                                      incremental=False, force=True)
            else:
                cinder.backups.create(volume_id, container, name, description,
                                      incremental=incremental, force=True)
        else:
            cinder.backups.create(volume_id, container, name, description,
                                  incremental=incremental, force=True)
